import time
import json
import boto3
import re
import traceback
import logging
import sys

from botocore.exceptions import ClientError
from awslambdatools.utility_functions import convertir_a_porcentaje

# Introducimos un logger para trazabilidad de errores
logger = logging.getLogger()
logger.setLevel(logging.INFO)

bedrock_runtime = boto3.client(
    service_name='bedrock-runtime',
    region_name='eu-central-1'
)

def claude_call(prompt, model = "anthropic.claude-instant-v1"):
    """
    Hace un llamamiento a un modelo Bedrock tipo Claude y obtiene la respuesta más otros metadatos

    Args:
        prompt (str): Prompt para el modelo generativo, al partir del cual generará una respuesta.
        model (str): Nombre del ID del modelo. Más info: https://docs.aws.amazon.com/bedrock/latest/userguide/model-ids-arns.html

    Returns:
        dict: Un diccionario con los siguientes campos:
        {
            Fecha (str): Fecha de la ejecución del modelo Bedrock.
            Respuesta (str): Respuesta del modelo Bedrock.
            N_tokens_input (int: Nº de tokens del prompt.
            N_tokens_output (int): Nº de tokens de la respuesta devuelta.
            Tiempo de ejecución (float): Tiempo de respuesta de la llamada al modelo.
        }
    """

    # Chequeo de la calidad del prompt
    if type(prompt) != str:
        raise TypeError(f"El prompt ha de ser de tipo str, no de tipo {type(prompt)}")
    
    if re.search("^\n*Human:", prompt) is None:
        raise ValueError(f"El prompt debe incluir 'Human:' al principio")

    # Añade un 'Assistant:' al final del prompt porque claude necesita eso para poder ejecutarse
    prompt_extended = prompt + 'Assistant:'

    claude_dict_body = {
        "prompt": prompt_extended,
        "temperature": 0.1,
        "top_p": 1,
        "top_k": 500,
        "max_tokens_to_sample": 4000
    }

    inicio = time.time()

    intentos = 0
    max_intentos = 5  # Puedes ajustar esto según tus necesidades
    espera_base = 5  # Puedes ajustar esto según tus necesidades

    while intentos < max_intentos:
        try:
            logger.info(f"Empieza la ejecución de Claude: Intento número {intentos}")
            claude_response = bedrock_runtime.invoke_model(modelId = model,
                                                    contentType = "application/json",
                                                    accept = "*/*",
                                                    body = json.dumps(claude_dict_body))
            logger.info(f"Termina el intento {intentos}")
            final = time.time()
            tiempo_total = round(final - inicio, 3)
            logger.info(f"Claude ha tardado unos {tiempo_total}")

            # Importación de la respuesta de formato JSON a un diccionario de Python
            claude_response_body = json.loads(claude_response.get('body').read())
            
            # Extraemos justo el output del modelo, sin metadatos
            respuesta = claude_response_body['completion']
            logger.info(f"Respuesta devuelta por invoke_model: {respuesta}")
            
            respuesta_dict = {
                            'Fecha': claude_response['ResponseMetadata']['HTTPHeaders']['date'],
                            'Respuesta': respuesta,
                            'N_tokens_input': len(prompt),
                            'N_tokens_output': round(len(respuesta) / 6),
                            'Tiempo_de_ejecucion': tiempo_total
                            }
            
            logger.info(f"Claude ha tardado unos {tiempo_total} segundos y {intentos + 1} intento/s en responder al prompt.")
            logger.info(f"Contenido del output de claude_call:{json.dumps(respuesta_dict)}")
            break
        


        except ClientError as e:
            # Excepción si hay un error de Throttling
            if e.response['Error']['Code'] == 'ThrottlingException':
                # Manejar la excepción de Throttling
                espera = (2 ** intentos) * espera_base
                logger.info(f"Throttling detectado. Esperando {espera} segundos antes de reintentar.")
                time.sleep(espera)
                intentos += 1

            # Excepción si hay un error de TimeOut
            elif e.response['Error']['Code'] == 'ModelTimeoutException':
                # Manejar la excepción de Throttling
                espera = (2 ** intentos) * espera_base
                logger.info(f"Modeltimeout detectado. Esperando {espera} segundos antes de reintentar.")
                time.sleep(espera)
                intentos += 1

            # Excepción si se produce un ValidationException
            elif e.response['Error']['Code'] == 'ValidationException':
                exception_type, exception_value, exception_traceback = sys.exc_info()
                traceback_string = traceback.format_exception(exception_type, exception_value, exception_traceback)
                err_msg = json.dumps({
                    "errorType": exception_type.__name__,
                    "errorMessage": str(exception_value),
                    "stackTrace": traceback_string,
                    "prompt": prompt,
                    "modelo_bedrock": model
                })
                logger.error(err_msg)
                raise ValueError('invoke_model ha recibido un parámetro erróneo')


            else:
                exception_type, exception_value, exception_traceback = sys.exc_info()
                traceback_string = traceback.format_exception(exception_type, exception_value, exception_traceback)
                err_msg = json.dumps({
                    "errorType": exception_type.__name__,
                    "errorMessage": str(exception_value),
                    "stackTrace": traceback_string,
                    "prompt": prompt,
                    "modelo_bedrock": model
                })
                logger.error(err_msg)
    return respuesta_dict


def bedrock_check(json_string):
    """
    Lee la respuesta del modelo bedrock, la transforma en un diccionario python y
    adapta el formato de cada variable para MySQL.

    Args:
        json_string (str): Un str con forma de objeto JSON. Ha de poder ser leído por json.json.loads()

    Returns:
        dict o list: Diccionario con los campos solicitados de bedrock si sólo hay una variable, y una 
                    lista si se está respondiendo a más de una variable.
    """
   


    # Meter un bucle linea por linea e vez de eliminar newlines
    json_razon = ""
    pattern_razon = '(?<=RAZON_RESULTADO": ").+\\s*\\w*' 

    for linea in json_string.splitlines():

        # Si la línea contiene RAZON_RESULTADO
        if re.search("RAZON_RESULTADO", linea) is not None:
            
            
            # Patrón de búsqueda para encontrar ":"
            patron_busqueda = r":(.*)$"
            
            # Realizar la búsqueda usando re.search
            resultado_busqueda = re.search(patron_busqueda, linea)
            
            # Verificar si se encontró una coincidencia
            if resultado_busqueda:
                # Obtener el contenido después de los dos puntos en ls_resultado
                razon_valor = resultado_busqueda.group(1)
                
                print ("++++++++++++++++++++++++++++++++++++")
                print ("++++++++RAZON DEL RESULTADO+++++++++")
                print (razon_valor)
                print ("++++++++++++++++++++++++++++++++++++")
                print ("++++++++++++++++++++++++++++++++++++")
            
            
            #resultado_busqueda = razon_valor = re.search(pattern_razon, linea)

            # Verificar si se encontró una coincidencia
            #if resultado_busqueda:
            #    razon_valor = resultado_busqueda.group(0)
                
                # Captura lo que hay después de RAZON RESULTADO...
                #razon_valor = re.search(pattern_razon, linea)[0]
                
                # ... y quítale las comillas dobles o simples
                razon_valor_cleaned = re.sub("\'|\"","",razon_valor)
                razon_valor_cleaned= re.sub(r'\\', "", razon_valor_cleaned)
                
                
                
    
                if "}" in linea:
                    # Añádelo al resto del texto y un símbolo }
                    json_razon +=  "        " + f"\"RAZON_RESULTADO\": \"{razon_valor_cleaned}\"" + "}" + '\n'
                else:
                    # Añádelo al resto del texto
                    json_razon +=  "        " + f"\"RAZON_RESULTADO\": \"{razon_valor_cleaned}\"" + '\n'
                




        else:
            json_razon += linea + '\n'
            
    print ("-----------------------------------------------------------------------------------------")
    print ("--------------------------------ANTES DE ELIMINAR RETONOS DE CARRO-----------------------")
    
    print (json_razon)
    print ("-----------------------------------------------------------------------------------------")
    print ("-----------------------------------------------------------------------------------------")

    logger.info(f'Resultado del json_string después de procesar RAZON_RESULTADO: {json_razon}')

    # Definimos un patrón que extrae el JSON de la respuesta de bedrock
    pattern_oldest = '[\\\[|{].*[\\]|}]'
    pattern_old = '\\[*[\\s\\[|{\\s].*[\\s\]|}\\s]\]*'
    pattern = '\\[*\\s*{.+}\\s*\]*'
    
    # Elimina las newlines del texto para facilitar la extracción del patrón
    json_no_newline = json_razon.replace("\n", "")

    # Si la respuesta viene con algún tipo de texto adicional además del json, extrar el json.
    
    
    
    
    # En caso contrario, devuelve error
    if re.search(pattern, json_no_newline) is not None:
        json_clean = re.search(pattern, json_no_newline)[0]
    else:
        raise ValueError("La respuesta de Bedrock no contiene ningún patrón JSON.")
           
    # A veces, RAZON_RESULTADO viene con comillas dobles entre comillas simples, 
    # produciendo un error de decodificación JSON. Las eliminamos con el siguiente patrón:
    # pattern_less_quotes = '(?<=\')"(?=.)|(?<=.)"(?=\')'
    
    # pattern_less_quotes = '(?<=RAZON_RESULTADO": ").+\\s*\\w'
    # json_clean = re.sub(pattern_less_quotes, "", json_clean)

    # pattern_less_quotes = '(?<=\')"(?=.)|(?<=.)"(?=\')'
    # json_clean = re.sub(pattern_less_quotes, "", json_clean)
    
    # Si además viene con comillas simples, sustituírlas por unas dobles
    # json_clean = re.sub("\"",'\'', json_clean)
    # Si además viene con comillas simples, sustituírlas por unas dobles
    json_clean = re.sub("\'",'\"', json_clean)
    try:
        razon_valor_cleaned= re.sub("\'", "\"", json_razon)
        
        print ("-----------------------------------------------------------------------------------------")
        print ("--------------------------------ANTES DE PASAR A JSON-----------------------")
        
        print (json_razon)
        print ("-----------------------------------------------------------------------------------------")
        print ("-----------------------------------------------------------------------------------------")        
        # Carga el string en un diccionario o en una lista de diccionarios
        dict_list = json.loads(json_clean)

        # Si es únicamente un diccionario, se crea una lista de un sólo diccionario
        if type(dict_list) == dict:
            dict_list = [json.loads(json_clean)]

        return_list = []
        print ("-----------------------------------------------------------------------------------------")
        print ("--------------------------------DESPUES DE PASAR A JSON-----------------------")
        
        print (dict_list)
        print ("-----------------------------------------------------------------------------------------")
        print ("-----------------------------------------------------------------------------------------")   
        
        for dict_index in dict_list :
    
            texto_clean = ''
            criterio_clean = ''
    
            # Asegura que ID es un integer
            id_clean = int(dict_index['ID'])
            print (f"***id_clean: {id_clean}")
    
            # Asegura que PORCENTAJE_DEL_RESULTADO devuelve un float adecuado comprendIDo entre 0 y 1
            porcentaje_clean = dict_index['PORCENTAJE_DEL_RESULTADO']
            porcentaje_clean = convertir_a_porcentaje(porcentaje_clean)
    
            # Asegura que CUMPLE CRITERIO devuelve estrictamente SI o NO, si existe
            if 'CUMPLE_CRITERIO' in dict_index:
                criterio_clean = dict_index['CUMPLE_CRITERIO'].upper()
    
            print (f"***Se mira si el texto de salida es un string")

            # Asegura que TEXTO SALIDA es un str, si existe, y lo mete en texto clean
            if 'TEXTO_SALIDA' in dict_index:
                if type(dict_index['TEXTO_SALIDA']) != str:
                    raise TypeError(f'TEXTO_SALIDA no es un str, si no de tipo {type(dict_index["TEXTO_SALIDA"])}')
                criterio_clean = dict_index['TEXTO_SALIDA']
            print (f"***texto_clean: {texto_clean}")    
    
            print (f"***MIRA SI RAZON_RESULTADO ES UN STRING")    
            # Asegura que RAZON_RESULTADO es un str, si existe
            if 'RAZON_RESULTADO' in dict_index:
                if type(dict_index['RAZON_RESULTADO']) != str:
                    raise TypeError(f'RAZON_RESULTADO no es un str, si no de tipo {type(dict_index["RAZON_RESULTADO"])}')
            
            print (f"***BEDROCK_CHECK devuelve ID: {id_clean} - CUMPLE_CRITERIO {criterio_clean} - TEXTO_SALIDA {texto_clean} - PORCENTAJE_DEL_RESULTADO {porcentaje_clean} - RAZON_RESULTADO {dict_index['RAZON_RESULTADO']} ")    
            # Devuelve el diccionario revisado
            return_list.append({
                    "ID":id_clean,
                    "CUMPLE_CRITERIO":criterio_clean,
                    "PORCENTAJE_DEL_RESULTADO": porcentaje_clean,
                    "RAZON_RESULTADO":dict_index['RAZON_RESULTADO'].strip(' ')
                }
            )
            
        if len(return_list) == 1:
            return return_list[0]
        else:
            return return_list

    except json.decoder.JSONDecodeError as json_err:
        logger.error(
                json.dumps(
                {
                'Mensaje':'JSONDecodeError: La respuesta no puede ser parseada como JSON. Devolvemos None en string',
                'string_de_error': json_clean,
                'errorType': 'JSONDecodeError'
                }
                )
            )
        # Se devuelve un None con la intención de que se vuelva a invocar Bedrock para preguntar
        # lo mismo y obtener una pregunta mejor
        return 'None'

    except Exception as exp:
        exception_type, exception_value, exception_traceback = sys.exc_info()
        traceback_string = traceback.format_exception(exception_type, exception_value, exception_traceback)
        err_msg = json.dumps({
            "json_string": json_string,
            "errorType": exception_type.__name__,
            "errorMessage": str(exception_value),
            "stackTrace": traceback_string
        })
        logger.error(err_msg)


def bedrock_iterator(prompt, model = "anthropic.claude-instant-v1", max_intentos = 8):
    """
    Wrapper que se ocupa de llamar a claude_call más de una vez si
    su respuesta en formato JSON tiene errores en el texto, tales
    como comillas dobles entre palabras. Comprueba la calidad de la
    respuesta mediante bedrock_check.

    Args:
        prompt (str): Prompt para el modelo generativo, al partir del cual generará una respuesta.
        model (str): Nombre del ID del modelo. Más info: https://docs.aws.amazon.com/bedrock/latest/userguide/model-ids-arns.html
        max_intentos (int): Intentos máximos para volver a llamar a Bedrock en caso de que la respuesta no sea satisfactoria

    Returns:
        dict: Un diccionario con los siguientes campos:
        -  Fecha (str): Fecha de la ejecución del modelo Bedrock.
        -  N_intento (int): Intento en el que se obtuvo la respuesta final.
        -  Respuesta_checked (dict o list): Respuesta del modelo Bedrock después de N número de intentos.
        -  N_tokens_input (int: Nº de tokens del prompt.
        -  N_tokens_output (int): Nº de tokens de la respuesta devuelta.
        -  Tiempo de ejecución (float): Tiempo de respuesta de la llamada al modelo.
    """
    # Comprobamos los tipos de los inputs
    if type(prompt) != str:
        raise TypeError(f'El mensaje no es de tipo str, si no {type(prompt)}')
    
    if type(model) != str:
        raise TypeError(f'El mensaje no es de tipo str, si no {type(model)}')
    
    if type(max_intentos) != int:
        raise TypeError(f'El mensaje no es de tipo str, si no {type(prompt)}')

    # Inicializamos intentos
    intento = 0

    # Inicio del loop
    while intento < max_intentos:
        try:
            logger.info(f'Llamada a Bedrock. Intento número:{intento + 1}')
            # Llamada a un modelo bedrock
            claude_answer = claude_call(prompt = prompt, model = model)

            # Verificación de calidad de la respuesta
            claude_checked_answer = bedrock_check(claude_answer['Respuesta'])

            # Si bedrock_check devuelve None, inicia el loop de nuevo. En caso
            # contrario, fin del loop.
            if claude_checked_answer == 'None':
                logger.info(
                    json.dumps(     
                                {
                                "Message":'Respuesta insatisfactoria: Regreso al inicio del loop',
                                "Pregunta": prompt,
                                "Respuesta_erronea": claude_answer['Respuesta']
                                }
                            )
                    )
                intento += 1
                continue
            else:
                logger.info(f'Fin del loop: Finaliza llamadas a bedrock después de {intento} intentos')
                break   

        except Exception as e:
                exception_type, exception_value, exception_traceback = sys.exc_info()
                traceback_string = traceback.format_exception(exception_type, exception_value, exception_traceback)
                err_msg = json.dumps({
                    "errorType": exception_type.__name__,
                    "errorMessage": str(exception_value),
                    "stackTrace": traceback_string,
                    "prompt": prompt,
                    "modelo_bedrock": model
                })
                logger.error(err_msg)
                raise ValueError("Ha devuelto un error la función")
    

    # Devolvemos el mismo diccionario que claude_call, pero con dos elementos adicionales:
    # - El intento en el que se obtuvo la respuesta buena
    # - La respuesta devuelta por bedrock_check, que puede ser o un diccionario o una lista
    claude_answer['N_intento'] = intento

    # Si se alcanza el número máximo de intentos, notificar y devolver la respuesta como inválida
    if intento == max_intentos:
        claude_answer['Respuesta_checked'] = f"Respuesta insatisfactoria después de {intento} intentos."
    else :
        claude_answer['Respuesta_checked'] = claude_checked_answer
                
    # Devolvemos el diccionario final
    return claude_answer 
